from dataclasses import dataclass
from sqlalchemy import func
from sqlalchemy.orm import Query
from typing import Type

from trailblazer.store.models import Model, Job


@dataclass
class BaseHandler_2:
    """All models in one base class."""

    def __init__(self):
        pass

    def get_query(self, table: Type[Model]) -> Query:
        """Return a query for the given table."""
        return self.query(table)

    def get_job_query_with_name_and_count_labels(self) -> Query:
        """Return a Job query with a name label and a count with a label."""
        return self.session.query(
            Job.name.label("name"),
            func.count(Job.id).label("count"),
        )



from pathlib import Path

from trailblazer.io.yaml import read_yaml


def test_get_content_from_file(example_yaml_path: Path):
    """
    Tests reading a YAML file.
    """
    # GIVEN a YAML file

    # WHEN reading the YAML file
    yaml_content: dict = read_yaml(file_path=example_yaml_path)

    # Then assert a dict is returned
    assert isinstance(yaml_content, dict)

    # Then assert content is unchanged
    assert yaml_content == {
        "users": [
            {"name": "Paul Anderson"},
        ],
    }



from typing import List

from tests.mocks.store_mock import MockStore
from trailblazer.store.models import Analysis


def test_delete_analysis_jobs(analysis_store: MockStore, tower_jobs: List[dict], case_id: str):
    """Test jobs are successfully deleted."""

    # GIVEN an analysis without failed jobs
    analysis: Analysis = analysis_store.get_latest_analysis(case_id=case_id)
    assert not analysis.failed_jobs

    # WHEN jobs are updated
    analysis_store.update_analysis_jobs(analysis=analysis, jobs=tower_jobs[:2])

    assert analysis.failed_jobs

    # WHEN jobs are deleted
    analysis_store.delete_analysis_jobs(analysis=analysis)

    # THEN analysis object should have no jobs
    assert not analysis.failed_jobs



from enum import Enum
from typing import Callable, List, Optional

from sqlalchemy.orm import Query

from trailblazer.store.models import Analysis


def filter_analyses_by_id(analyses: Query, analysis_id: int, **kwargs) -> Query:
    """Filter analyses by database entry id."""
    return analyses.filter(Analysis.id == analysis_id)


class AnalysisFilter(Enum):
    """Define Analysis filter functions."""

    FILTER_BY_ID: Callable = filter_analyses_by_id


def apply_analysis_filter(
    analyses: Query,
    filter_functions: List[Callable],
    analysis_id: Optional[int] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        analyses: Query = function(
            analyses=analyses,
            analysis_id=analysis_id,
        )
    return analyses



from trailblazer.apps.slurm.api import reformat_squeue_result_job_step
from trailblazer.apps.slurm.models import SqueueResult
from typing import List

from trailblazer.store.base import BaseHandler_2
from trailblazer.store.models import User, Analysis, Job


class UpdateHandler(BaseHandler_2):
    """Class for updating items in the database."""

    def add_user(self, name: str, email: str) -> User:
        """Add a new user to the database."""
        new_user: User = User(email=email, name=name)
        self.add_commit(new_user)
        return new_user

    def update_analysis_jobs(self, analysis: Analysis, jobs: List[dict]) -> None:
        """Update jobs in the analysis."""
        # failed_jobs is misnamed and actually contains all jobs irrespective of status
        analysis.failed_jobs = [Job(**job) for job in jobs]
        self.commit()

    def update_user_is_archived(self, user: User, archive: bool = True) -> None:
        """Update is archived for a user in the database."""
        user.is_archived = archive
        self.commit()

    def update_analysis_jobs_from_slurm_jobs(
        self, analysis: Analysis, squeue_result: SqueueResult
    ) -> None:
        """Update analysis failed jobs from supplied squeue results."""
        if len(squeue_result.jobs) == 0:
            return
        for job in squeue_result.jobs:
            job.step = reformat_squeue_result_job_step(
                data_analysis=analysis.data_analysis, job_step=job.step
            )

        for job_obj in analysis.failed_jobs:
            job_obj.delete()
        self.commit()
        analysis.failed_jobs = [
            self.Job(
                analysis_id=analysis.id,
                slurm_id=job.id,
                name=job.step,
                status=job.status,
                started_at=job.started_at,
                elapsed=job.time_elapsed,
            )
            for job in squeue_result.jobs
        ]
        self.commit()



"""Store backend in Trailblazer."""
import datetime as dt
import logging
import subprocess
from pathlib import Path
from typing import Any, Optional

import alchy
import sqlalchemy as sqa
from alchy import Query

from trailblazer.apps.slurm.api import get_squeue_result, get_current_analysis_status
from trailblazer.apps.slurm.models import SqueueResult
from trailblazer.apps.tower.api import TowerAPI
from trailblazer.constants import (
    COMPLETED_STATUS,
    FAILED_STATUS,
    ONGOING_STATUSES,
    SLURM_ACTIVE_CATEGORIES,
    STARTED_STATUSES,
    STATUS_OPTIONS,
    FileFormat,
    TrailblazerStatus,
    WorkflowManager,
    SlurmJobStatus,
)
from trailblazer.exc import TowerRequirementsError, TrailblazerError
from trailblazer.io.controller import ReadFile
from trailblazer.store.core import CoreHandler
from trailblazer.store.models import Analysis, Job, Model

LOG = logging.getLogger(__name__)


class BaseHandler(CoreHandler):
    Analysis = Analysis
    Job = Job

    def setup(self):
        self.create_all()

    def get_analysis(self, case_id: str, started_at: dt.datetime, status: str) -> Analysis:
        """
        used in LOG
        Find a single analysis."""
        query = self.Analysis.query.filter_by(family=case_id, started_at=started_at, status=status)
        return query.first()

    def analyses(
        self,
        case_id: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
        family: str = None,
        data_analysis: str = None,
        comment: str = None,
    ) -> Query:
        """
        used by REST +> CG
        Fetch analyses from the database."""
        if not case_id:
            case_id = family

        analysis_query = self.Analysis.query
        if case_id:
            analysis_query = analysis_query.filter_by(family=case_id)
        if query:  # to be deprecated
            analysis_query = analysis_query.filter(
                sqa.or_(
                    self.Analysis.family.ilike(f"%{query}%"),
                    self.Analysis.status.ilike(f"%{query}%"),
                    self.Analysis.data_analysis.ilike(f"%{query}%"),
                    self.Analysis.comment.ilike(f"%{query}%"),
                )
            )
        if status:
            analysis_query = analysis_query.filter_by(status=status)
        if isinstance(deleted, bool):
            analysis_query = analysis_query.filter_by(is_deleted=deleted)
        if temp:
            analysis_query = analysis_query.filter(self.Analysis.status.in_(ONGOING_STATUSES))
        if before:
            analysis_query = analysis_query.filter(self.Analysis.started_at < before)
        if is_visible is not None:
            analysis_query = analysis_query.filter_by(is_visible=is_visible)
        if data_analysis:
            analysis_query = analysis_query.filter(
                Analysis.data_analysis.ilike(f"%{data_analysis}%")
            )
        if comment:
            analysis_query = analysis_query.filter(Analysis.comment.ilike(f"%{comment}%"))

        return analysis_query.order_by(self.Analysis.started_at.desc())

    def get_latest_analysis(self, case_id: str) -> Optional[Analysis]:
        return self.analyses(case_id=case_id).first()

    def get_latest_analysis_status(self, case_id: str) -> Optional[str]:
        """Get the latest analysis status for a case_id"""
        latest_analysis = self.get_latest_analysis(case_id=case_id)
        if latest_analysis:
            return latest_analysis.status

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        """Check if the latest analysis is ongoing for a case_id"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status in ONGOING_STATUSES

    def is_latest_analysis_failed(self, case_id: str) -> bool:
        """Check if the latest analysis is failed for a case_id"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status == FAILED_STATUS

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        """Check if the latest analysis is completed for a case_id"""
        latest_analysis = self.analyses(case_id=case_id).first()
        return bool(latest_analysis and latest_analysis.status == COMPLETED_STATUS)

    def has_latest_analysis_started(self, case_id: str) -> bool:
        """Check if analysis has started"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status in STARTED_STATUSES

    def mark_analyses_deleted(self, case_id: str) -> Query:
        """mark analyses connected to a case as deleted"""
        old_analyses = self.analyses(case_id=case_id)
        if old_analyses.count() > 0:
            for old_analysis in old_analyses:
                old_analysis.is_deleted = True
            self.commit()
        return old_analyses

    def set_analysis_completed(self, analysis_id: int) -> None:
        """Set an analysis status to completed."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        analysis.status = TrailblazerStatus.COMPLETED
        self.commit()
        LOG.info(f"{analysis.family} - status set to {TrailblazerStatus.COMPLETED}")

    def set_analysis_uploaded(self, case_id: str, uploaded_at: dt.datetime) -> None:
        """Setting analysis uploaded at."""
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.uploaded_at: dt.datetime = uploaded_at
        self.commit()

    def set_analysis_status(self, case_id: str, status: str):
        """Setting analysis status."""
        status = status.lower()
        if status not in set(STATUS_OPTIONS):
            raise ValueError(f"Invalid status. Allowed values are: {STATUS_OPTIONS}")
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.status: str = status
        self.commit()
        LOG.info(f"{analysis_obj.family} - Status set to {status.upper()}")

    def add_comment(self, case_id: str, comment: str):
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.comment: str = (
            " ".join([analysis_obj.comment, comment]) if analysis_obj.comment else comment
        )

        self.commit()

        LOG.info(f"Adding comment {comment} to analysis {analysis_obj.family}")

    def delete_analysis(self, analysis_id: int, force: bool = False) -> None:
        """Delete the analysis output."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        if not analysis:
            raise TrailblazerError("Analysis not found")

        if not force and analysis.status in ONGOING_STATUSES:
            raise TrailblazerError(
                f"Analysis for {analysis.family} is currently running! Use --force flag to delete anyway."
            )
        LOG.info(f"Deleting analysis {analysis_id} for case {analysis.family}")
        analysis.delete()
        self.commit()

    @staticmethod
    def query_slurm(job_id_file: str, case_id: str, ssh: bool) -> Any:
        """Args:
        job_id_file: Path to slurm id .YAML file as string
        case_id: Unique internal case identifier which is expected to by the only item in the .YAML dict
        ssh : Whether the request is executed from hasta or clinical-db"""
        job_id: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=Path(job_id_file)
        )
        submitted_job_ids = job_id.get(next(iter(job_id)))
        job_ids_string = ",".join(map(str, submitted_job_ids))
        if ssh:
            return (
                subprocess.check_output(
                    [
                        "ssh",
                        "hiseq.clinical@hasta.scilifelab.se",
                        "squeue",
                        "-j",
                        job_ids_string,
                        "--states=all",
                        "-o",
                        "%A,%j,%T,%l,%M,%S",
                    ],
                    universal_newlines=True,
                )
                .decode("utf-8")
                .strip()
                .replace("//n", "/n")
            )
        else:
            return subprocess.check_output(
                [
                    "squeue",
                    "-j",
                    job_ids_string,
                    "--states=all",
                    "-o",
                    "%A,%j,%T,%l,%M,%S",
                ]
            ).decode("utf-8")

    def update_ongoing_analyses(self, ssh: bool = False) -> None:
        """Iterate over all analysis with ongoing status and query SLURM for current progress."""
        ongoing_analyses = self.analyses(temp=True)
        for analysis_obj in ongoing_analyses:
            try:
                self.update_run_status(analysis_id=analysis_obj.id, ssh=ssh)
            except Exception as error:
                LOG.error(
                    f"Failed to update {analysis_obj.family} - {analysis_obj.id}: {type(error).__name__}"
                )

    def update_run_status(self, analysis_id: int, ssh: bool = False) -> None:
        """Query entries related to given analysis, and update the Trailblazer database."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        if not analysis:
            LOG.warning(f"Analysis {analysis_id} not found!")
            return
        if analysis.workflow_manager == WorkflowManager.TOWER.value:
            self.update_tower_run_status(analysis_id=analysis_id)
        elif analysis.workflow_manager == WorkflowManager.SLURM.value:
            self.update_analysis_from_slurm_run_status(analysis_id=analysis_id, ssh=ssh)

    def update_analysis_from_slurm_run_status(self, analysis_id: int, ssh: bool = False) -> None:
        """Query slurm for entries related to given analysis, and update the analysis in the database."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        try:
            squeue_result: SqueueResult = get_squeue_result(
                squeue_response=self.query_slurm(
                    job_id_file=analysis.config_path, case_id=analysis.family, ssh=ssh
                )
            )
            self.update_analysis_jobs_from_slurm_jobs(
                analysis=analysis, squeue_result=squeue_result
            )
            LOG.info(f"Status in SLURM: {analysis.family} - {analysis_id}")
            LOG.debug(squeue_result.jobs)
            analysis.progress = squeue_result.jobs_status_distribution.get(
                SlurmJobStatus.COMPLETED, 0.0
            )
            analysis.status = get_current_analysis_status(
                jobs_status_distribution=squeue_result.jobs_status_distribution
            )
            LOG.info(f"Updated status {analysis.family} - {analysis.id}: {analysis.status} ")
            self.commit()

            analysis.logged_at = dt.datetime.now()
        except Exception as exception:
            LOG.error(
                f"Error updating analysis for: case - {analysis.family} : {exception.__class__.__name__}"
            )
            analysis.status = TrailblazerStatus.ERROR
            self.commit()

    @staticmethod
    def query_tower(config_file: str, case_id: str) -> TowerAPI:
        """Parse a config file to extract a NF Tower workflow ID and return a TowerAPI.
        Currently only one tower ID is supported."""
        workflow_id: int = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=Path(config_file)
        ).get(case_id)[-1]
        tower_api = TowerAPI(workflow_id=workflow_id)
        if not tower_api.tower_client.meets_requirements:
            raise TowerRequirementsError
        return tower_api

    def update_tower_run_status(self, analysis_id: int) -> None:
        """Query tower for entries related to given analysis, and update the Trailblazer database."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        tower_api: TowerAPI = self.query_tower(
            config_file=analysis.config_path, case_id=analysis.family
        )

        try:
            LOG.info(
                f"Status in Tower: {analysis.family} - {analysis_id} - {tower_api.workflow_id}"
            )
            analysis.status: str = tower_api.status
            analysis.progress: int = tower_api.progress
            analysis.logged_at: dt.datetime = dt.datetime.now()
            self.delete_analysis_jobs(analysis=analysis)
            self.update_analysis_jobs(
                analysis=analysis, jobs=tower_api.get_jobs(analysis_id=analysis.id)
            )
            self.commit()
            LOG.info(f"Updated status {analysis.family} - {analysis.id}: {analysis.status} ")
        except Exception as error:
            LOG.error(f"Error logging case - {analysis.family} :  {type(error).__name__}")
            analysis.status: str = TrailblazerStatus.ERROR
            self.commit()

    @staticmethod
    def cancel_slurm_job(slurm_id: int, ssh: bool = False) -> None:
        """Cancel slurm job by slurm job ID"""
        if ssh:
            subprocess.Popen(
                ["ssh", "hiseq.clinical@hasta.scilifelab.se", "scancel", str(slurm_id)]
            )
        else:
            subprocess.Popen(["scancel", str(slurm_id)])

    def cancel_analysis(self, analysis_id: int, email: str = None, ssh: bool = False) -> None:
        """Cancel all ongoing slurm jobs associated with the analysis, and set job status to canceled"""
        analysis: Analysis = self.get_analysis_with_id(analysis_id=analysis_id)
        if not analysis:
            raise TrailblazerError(f"Analysis {analysis_id} does not exist")

        if analysis.status not in ONGOING_STATUSES:
            raise TrailblazerError(f"Analysis {analysis_id} is not running")

        for job_obj in analysis.failed_jobs:
            if job_obj.status in SLURM_ACTIVE_CATEGORIES:
                LOG.info(f"Cancelling job {job_obj.slurm_id} - {job_obj.name}")
                self.cancel_slurm_job(job_obj.slurm_id, ssh=ssh)
        LOG.info(
            f"Case {analysis.family} - Analysis {analysis_id}: all ongoing jobs cancelled successfully!"
        )
        self.update_run_status(analysis_id=analysis_id)
        analysis.status = TrailblazerStatus.CANCELLED
        analysis.comment = (
            f"Analysis cancelled manually by user:"
            f" {(self.get_user(email=email).name if self.get_user(email=email) else (email or 'Unknown'))}!"
        )
        self.commit()


class Store(alchy.Manager, BaseHandler):
    def __init__(self, uri: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model)



import datetime
import multiprocessing
import os
from typing import Dict, Mapping, List, Union

from dateutil.parser import parse as parse_datestr
from flask import Blueprint, Response, abort, g, jsonify, make_response, request
from google.auth import jwt

from trailblazer.constants import TrailblazerStatus, ONE_MONTH_IN_DAYS
from trailblazer.server.ext import store
from trailblazer.store.models import Info, User, Analysis
from trailblazer.utils.date import get_date_days_ago

blueprint = Blueprint("api", __name__, url_prefix="/api/v1")


def stringify_timestamps(data: dict) -> Dict[str, str]:
    """Convert datetime into string before dumping in order to avoid information loss"""
    for key, val in data.items():
        if isinstance(val, datetime.datetime):
            data[key] = str(val)
    return data


@blueprint.before_request
def before_request():
    """Authentication that is run before processing requests to the application"""
    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), 204)
    if os.environ.get("SCOPE") == "DEVELOPMENT":
        return
    auth_header = request.headers.get("Authorization")
    if auth_header:
        jwt_token = auth_header.split("Bearer ")[-1]
    else:
        return abort(403, "no JWT token found on request")

    user_data: Mapping = jwt.decode(jwt_token, verify=False)
    user: User = store.get_user(email=user_data["email"], exclude_archived=True)
    if not user:
        return abort(403, f"{user_data['email']} doesn't have access")
    g.current_user = user


@blueprint.route("/analyses")
def analyses():
    """Display analyses."""
    per_page = int(request.args.get("per_page", 100))
    page = int(request.args.get("page", 1))
    query = store.analyses(
        status=request.args.get("status"),
        query=request.args.get("query"),
        is_visible=request.args.get("is_visible") == "true" or None,
    )

    query_page = query.paginate(page, per_page=per_page)
    data = []
    for analysis_obj in query_page.items:
        analysis_data = analysis_obj.to_dict()
        analysis_data["user"] = analysis_obj.user.to_dict() if analysis_obj.user else None
        analysis_data["failed_jobs"] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
        data.append(analysis_data)

    return jsonify(analyses=data)


@blueprint.route("/analyses/<int:analysis_id>", methods=["GET", "PUT"])
def analysis(analysis_id):
    """Display a single analysis."""
    analysis_obj = store.get_analysis_with_id(analysis_id=analysis_id)
    if analysis_obj is None:
        return abort(404)

    if request.method == "PUT":
        analysis_obj.update(request.json)
        store.commit()

    data = analysis_obj.to_dict()
    data["failed_jobs"] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
    data["user"] = analysis_obj.user.to_dict() if analysis_obj.user else None
    return jsonify(**data)


@blueprint.route("/info")
def info():
    """Display metadata about database."""
    info: Info = store.get_query(table=Info).first()
    return jsonify(**info.to_dict())


@blueprint.route("/me")
def me():
    """Return information about a logged in user."""
    return jsonify(**g.current_user.to_dict())


@blueprint.route("/aggregate/jobs")
def aggregate_jobs():
    """Return stats about failed jobs."""
    time_window: datetime = get_date_days_ago(
        days_ago=int(request.args.get("days_back", ONE_MONTH_IN_DAYS))
    )
    failed_jobs: List[Dict[str, Union[str, int]]] = store.get_nr_jobs_with_status_per_category(
        status=TrailblazerStatus.FAILED, since_when=time_window
    )
    return jsonify(jobs=failed_jobs)


@blueprint.route("/update-all")
def update_analyses():
    """Update all ongoing analysis by querying SLURM"""
    process = multiprocessing.Process(target=store.update_ongoing_analyses, kwargs={"ssh": True})
    process.start()
    return jsonify(f"Success! Trailblazer updated {datetime.datetime.now()}"), 201


@blueprint.route("/update/<int:analysis_id>", methods=["PUT"])
def update_analysis(analysis_id):
    """Update a specific analysis"""
    try:
        process = multiprocessing.Process(
            target=store.update_run_status, kwargs={"analysis_id": analysis_id, "ssh": True}
        )
        process.start()
        return jsonify("Success! Update request sent"), 201
    except Exception as e:
        return jsonify(f"Exception: {e}"), 409


@blueprint.route("/cancel/<int:analysis_id>", methods=["PUT"])
def cancel(analysis_id):
    """Cancel an analysis and all slurm jobs associated with it"""
    auth_header = request.headers.get("Authorization")
    jwt_token = auth_header.split("Bearer ")[-1]
    user_data = jwt.decode(jwt_token, verify=False)
    try:
        process = multiprocessing.Process(
            target=store.cancel_analysis,
            kwargs={"analysis_id": analysis_id, "email": user_data["email"], "ssh": True},
        )
        process.start()
        return jsonify("Success! Cancel request sent"), 201
    except Exception as e:
        return jsonify(f"Exception: {e}"), 409


@blueprint.route("/delete/<int:analysis_id>", methods=["PUT"])
def delete(analysis_id):
    """Cancel an analysis and all slurm jobs associated with it"""
    try:
        process = multiprocessing.Process(
            target=store.delete_analysis,
            kwargs={"analysis_id": analysis_id, "force": True},
        )
        process.start()
        return jsonify("Success! Delete request sent!"), 201
    except Exception as e:
        return jsonify(f"Exception: {e}"), 409


# CG REST INTERFACE ###
# ONLY POST routes which accept messages in specific format
# NOT for use with GUI (for now)


@blueprint.route("/query-analyses", methods=["POST"])
def post_query_analyses():
    """Return list of analyses matching the query terms"""
    content = request.json
    query_analyses = store.analyses(
        case_id=content.get("case_id"),
        query=content.get("query"),
        status=content.get("status"),
        deleted=content.get("deleted"),
        temp=content.get("temp"),
        before=parse_datestr(content.get("before")) if content.get("before") else None,
        is_visible=content.get("visible"),
        family=content.get("family"),
        data_analysis=content.get("data_analysis"),
    )
    data = [stringify_timestamps(analysis_obj.to_dict()) for analysis_obj in query_analyses]
    return jsonify(*data), 200


@blueprint.route("/get-latest-analysis", methods=["POST"])
def post_get_latest_analysis():
    """Return latest analysis entry for specified case"""
    content = request.json
    analysis_obj = store.get_latest_analysis(case_id=content.get("case_id"))
    if analysis_obj:
        data = stringify_timestamps(analysis_obj.to_dict())
        return jsonify(**data), 200
    return jsonify(None), 200


@blueprint.route("/find-analysis", methods=["POST"])
def post_find_analysis():
    """Find analysis using case_id, date, and status"""
    content = request.json
    analysis_obj = store.get_analysis(
        case_id=content.get("case_id"),
        started_at=parse_datestr(content.get("started_at")),
        status=content.get("status"),
    )
    if analysis_obj:
        data = stringify_timestamps(analysis_obj.to_dict())
        return jsonify(**data), 200
    return jsonify(None), 200


@blueprint.route("/delete-analysis", methods=["POST"])
def post_delete_analysis():
    """Delete analysis using analysis_id. If analysis is ongoing, error will be raised.
    To delete ongoing analysis, --force flag should also be passed.
    If an ongoing analysis is deleted in ths manner, all ongoing jobs will be cancelled"""
    content = request.json
    try:
        store.delete_analysis(analysis_id=content.get("analysis_id"), force=content.get("force"))
        return jsonify(None), 201
    except Exception as e:
        return jsonify(f"Exception: {e}"), 409


@blueprint.route("/mark-analyses-deleted", methods=["POST"])
def post_mark_analyses_deleted():
    """Mark all analysis belonging to a case deleted"""
    content = request.json
    old_analyses = store.mark_analyses_deleted(case_id=content.get("case_id"))
    data = [stringify_timestamps(analysis_obj.to_dict()) for analysis_obj in old_analyses]
    if data:
        return jsonify(*data), 201
    return jsonify(None), 201


@blueprint.route("/add-pending-analysis", methods=["POST"])
def post_add_pending_analysis():
    """Add new analysis with status: pending."""
    post_request: Response.json = request.json
    try:
        analysis: Analysis = store.add_pending_analysis(
            case_id=post_request.get("case_id"),
            email=post_request.get("email"),
            type=post_request.get("type"),
            config_path=post_request.get("config_path"),
            out_dir=post_request.get("out_dir"),
            priority=post_request.get("priority"),
            data_analysis=post_request.get("data_analysis"),
            ticket_id=post_request.get("ticket"),
            workflow_manager=post_request.get("workflow_manager"),
        )
        raw_analysis: dict = stringify_timestamps(analysis.to_dict())
        return jsonify(**raw_analysis), 201
    except Exception as exception:
        return jsonify(f"Exception: {exception}"), 409


@blueprint.route("/set-analysis-uploaded", methods=["PUT"])
def put_set_analysis_uploaded():
    content: Response.json = request.json

    try:
        store.set_analysis_uploaded(
            case_id=content.get("case_id"), uploaded_at=content.get("uploaded_at")
        )
        return jsonify("Success! Uploaded at request sent"), 201
    except Exception as error:
        return jsonify(f"Exception: {error}"), 409


@blueprint.route("/set-analysis-status", methods=["PUT"])
def put_set_analysis_status():
    content: Response.json = request.json

    try:
        store.set_analysis_status(case_id=content.get("case_id"), status=content.get("status"))
        return jsonify(f"Success! Analysis set to {content.get('status')} request sent"), 201
    except Exception as error:
        return jsonify(f"Exception: {error}"), 409


@blueprint.route("/add-comment", methods=["PUT"])
def put_add_comment():
    content: Response.json = request.json

    try:
        store.add_comment(case_id=content.get("case_id"), comment=content.get("comment"))
        return jsonify("Success! Adding comment request sent"), 201
    except Exception as error:
        return jsonify(f"Exception: {error}"), 409



from datetime import datetime, timedelta


def get_date_days_ago(days_ago: int) -> datetime:
    """Return the date that was number of 'days_ago'."""
    return datetime.now() - timedelta(days=days_ago)



from datetime import datetime

from trailblazer.utils.date import get_date_days_ago


def test_get_date_days_ago(timestamp_now: datetime):
    """Test returning a date corresponding to number of days ago."""

    # GIVEN days ago

    # WHEN calling the function
    date: datetime = get_date_days_ago(days_ago=1)

    # THEN assert the return should be a date
    assert isinstance(date, datetime)

    # Then the date returned should be less than then today
    assert date < timestamp_now



import pytest


@pytest.fixture(scope="session", name="squeue_stream_pending_job")
def fixture_squeue_stream_pending_job() -> str:
    """Return a squeue output stream."""
    return """JOBID,NAME,STATE,TIME_LIMIT,TIME,START_TIME
690990,gatk_genotypegvcfs6,PENDING,10:00:00,0:00,2020-10-22T11:42:00"""


@pytest.fixture(scope="session", name="squeue_stream_pending_job_not_started")
def fixture_squeue_stream_pending_job_not_started() -> str:
    """Return a squeue output stream for a ot started job."""
    return """JOBID,NAME,STATE,TIME_LIMIT,TIME,START_TIME
690990,gatk_genotypegvcfs6,PENDING,10:00:00,0:00,N/A"""



"""Model SLURM output."""
from datetime import datetime
from typing import List, Dict, Optional, Any

from pydantic import BaseModel, Field, validator

from trailblazer.constants import SlurmJobStatus, SlurmSqueueHeader
from trailblazer.exc import MissingSqueueOutput
from trailblazer.utils.datetime import (
    convert_days_to_minutes,
    convert_timestamp_to_minutes,
    get_datetime_from_timestamp,
)


class SqueueJob(BaseModel):
    """Model job meta data from squeue output."""

    id: int = Field(..., alias=SlurmSqueueHeader.JOBID.value)
    step: str = Field(..., alias=SlurmSqueueHeader.NAME.value)
    status: SlurmJobStatus = Field(..., alias=SlurmSqueueHeader.STATE.value)
    time_limit: str = Field(..., alias=SlurmSqueueHeader.TIME_LIMIT.value)
    time_elapsed: int = Field(..., alias=SlurmSqueueHeader.TIME.value)
    started_at: Optional[datetime] = Field(..., alias=SlurmSqueueHeader.START_TIME.value)

    @validator("status", always=True, pre=True)
    def convert_status_to_lower_case(cls, value: str) -> str:
        """Convert string to lower case."""
        return value.lower()

    @validator("time_elapsed", always=True, pre=True)
    def convert_time_elapsed_to_minutes(cls, value: str) -> int:
        """Convert squeue timestamp string into minutes."""
        raw_time_elapsed: str = value
        if not raw_time_elapsed or not isinstance(raw_time_elapsed, str):
            return 0
        timestamps: List[str] = []
        if "-" in raw_time_elapsed:
            timestamps: List[str] = raw_time_elapsed.split("-")
        day_nr: int = int(timestamps[0]) if timestamps else 0
        raw_timestamp: str = timestamps[1] if timestamps else raw_time_elapsed
        time_elapsed: datetime = get_datetime_from_timestamp(
            timestamp=raw_timestamp, datetime_formats=["%M:%S", "%H:%M:%S"]
        )
        return convert_timestamp_to_minutes(timestamp=time_elapsed) + convert_days_to_minutes(
            days_nr=day_nr
        )

    @validator("started_at", always=True, pre=True)
    def convert_started_at_to_datetime(cls, value: str) -> Optional[datetime]:
        """Convert started to datetime if string is in datetime format."""
        raw_started: str = value
        if isinstance(raw_started, str) and raw_started not in {"N/A"}:
            return datetime.strptime(raw_started, "%Y-%m-%dT%H:%M:%S")


class SqueueResult(BaseModel):
    """Model used to parse SLURM squeue output."""

    jobs: List[SqueueJob]
    jobs_status_distribution: Optional[Dict[str, float]]

    @validator("jobs_status_distribution", always=True)
    def set_jobs_status_distribution(
        cls, _: Optional[List[SqueueJob]], values: Dict[str, Any]
    ) -> Optional[Dict[str, float]]:
        """Set the fraction for each status present in the squeue jobs.
        Example {PENDING:0.33, RUNNING:0.33, FAILED: 0.33}."""
        jobs: List[SqueueJob] = values["jobs"]
        if not jobs:
            raise MissingSqueueOutput("No squeue output")
        status_distribution: Dict[str, float] = {}
        job_statuses: List[str] = [job.status for job in jobs]
        for job_status in set(job_statuses):
            status_distribution[job_status] = round(job_statuses.count(job_status) / len(jobs), 2)
        return status_distribution



from datetime import datetime
from typing import List, Dict, Union

from tests.mocks.store_mock import MockStore
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.models import User, Analysis
from tests.store.utils.store_helper import StoreHelpers


def test_get_analysis_with_id(analysis_store: MockStore):
    """Test getting an analysis when it exists in the database."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN accessing it by ID
    analysis: Analysis = analysis_store.get_analysis_with_id(analysis_id=existing_analysis.id)

    # THEN it should return the same analysis
    assert analysis == existing_analysis


def test_get_analysis_with_id_when_missing(analysis_store: MockStore):
    """Test getting an analysis when it does not exist in the database."""
    # GIVEN an id that doesn't exist
    missing_analysis_id = 12312423534

    # WHEN accessing the analysis
    analysis: Analysis = analysis_store.get_analysis_with_id(analysis_id=missing_analysis_id)

    # THEN it should return None
    assert not analysis


def test_get_nr_jobs_with_status_per_category(job_store: MockStore, timestamp_yesterday: datetime):
    """Test getting the number of failed jobs per category since a supplied date from the database."""
    # GIVEN a database with jobs

    # WHEN querying for failed users
    failed_jobs: List[Dict[str, Union[str, int]]] = job_store.get_nr_jobs_with_status_per_category(
        since_when=timestamp_yesterday, status=TrailblazerStatus.FAILED
    )

    # THEN failed jobs should be returned
    assert failed_jobs

    # THEN one job should be returned
    assert len(failed_jobs) == 1

    # THEN name and count should be defined
    assert failed_jobs[0].get("name") == "1"
    assert failed_jobs[0].get("count") == 1


def test_get_user(user_store: MockStore, user_email: str):
    """Test getting a user from the database."""
    # GIVEN a database with a user

    # WHEN querying for a user
    user: User = user_store.get_user(email=user_email)

    # THEN it should be returned
    assert user.email == user_email


def test_get_user_including_archived(user_store: MockStore, archived_user_email: str):
    """Test getting an archived user from the database."""
    # GIVEN a database with an archived user

    # WHEN querying for a user
    user: User = user_store.get_user(email=archived_user_email, exclude_archived=False)

    # THEN it should be returned
    assert user.email == archived_user_email


def test_get_user_including_archive_false(user_store: MockStore, archived_user_email: str):
    """Test getting an archived user from the database."""
    # GIVEN a database with an archived user

    # WHEN querying for a user
    user: User = user_store.get_user(email=archived_user_email, exclude_archived=True)

    # THEN no user should be returned
    assert not user


def test_get_user_when_non_existing(user_store: MockStore):
    """Test getting a non-existing user from the database."""
    # GIVEN a database with a user

    # WHEN querying for a user that doesn't exist
    user: User = user_store.get_user(email="this_is_a_made_up_email@fake_example.com")

    # THEN it should return as None
    assert user is None


def test_get_users(user_store: MockStore, user_email: str, username: str):
    """Test getting a user with username and email."""
    # GIVEN a database with a user

    # WHEN getting users
    users: List[User] = user_store.get_users(name=username, email=user_email)

    # THEN the user should be returned
    assert users[0].email == user_email


def test_get_users_including_archived(user_store: MockStore, user_email: str, username: str):
    """Test getting a user with username and email which are archived."""
    # GIVEN a database with a user and an archived user
    StoreHelpers.add_user(
        name=username, email="old.user@magnolia.com", store=user_store, is_archived=True
    )

    # WHEN getting users
    users: List[User] = user_store.get_users(name=username, exclude_archived=False)

    # THEN the two users should be returned
    assert len(users) == 2


def test_get_users_no_username(user_store: MockStore, user_email: str):
    """Test getting a user by email."""
    # GIVEN a database with a user

    # WHEN getting users
    users: List[User] = user_store.get_users(email=user_email)

    # THEN the user should be returned
    assert users[0].email == user_email



from enum import Enum
from typing import Callable, List, Optional

from sqlalchemy.orm import Query

from trailblazer.store.models import User


def filter_users_by_email(users: Query, email: str, **kwargs) -> Query:
    """Filter users with matching email."""
    return users.filter(User.email == email)


def filter_users_by_contains_email(users: Query, email: str, **kwargs) -> Query:
    """Filter users which contains email."""
    return users.filter(User.email.contains(email))


def filter_users_by_contains_name(users: Query, name: str, **kwargs) -> Query:
    """Filter users which contains name."""
    return users.filter(User.name.contains(name))


def filter_users_by_is_not_archived(users: Query, **kwargs) -> Query:
    """Filter users which are not archived."""
    return users.filter(User.is_archived.is_(False))


class UserFilter(Enum):
    """Define User filter functions."""

    FILTER_BY_CONTAINS_EMAIL: Callable = filter_users_by_contains_email
    FILTER_BY_CONTAINS_NAME: Callable = filter_users_by_contains_name
    FILTER_BY_EMAIL: Callable = filter_users_by_email
    FILTER_BY_IS_NOT_ARCHIVED: Callable = filter_users_by_is_not_archived


def apply_user_filter(
    users: Query,
    filter_functions: List[Callable],
    email: Optional[str] = None,
    name: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        users: Query = function(
            users=users,
            email=email,
            name=name,
        )
    return users



import logging
import sys
from pathlib import Path
from typing import List

import click
import coloredlogs
from dateutil.parser import parse as parse_date

import trailblazer
from trailblazer.cli.utils.user_helper import is_existing_user, is_user_archived
from trailblazer.constants import FileFormat
from trailblazer.environ import environ_email
from trailblazer.io.controller import ReadFile
from trailblazer.store.api import Store
from trailblazer.store.models import STATUS_OPTIONS, User

LOG = logging.getLogger(__name__)
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


@click.group()
@click.option("-c", "--config", type=click.File())
@click.option("-d", "--database", help="path/URI of the SQL database")
@click.option(
    "-l", "--log-level", type=click.Choice(LEVELS), default="INFO", help="lowest level to log at"
)
@click.option("--verbose", is_flag=True, help="Show full log information, time stamp etc")
@click.version_option(trailblazer.__version__, prog_name=trailblazer.__title__)
@click.pass_context
def base(
    context,
    config,
    database,
    log_level: str,
    verbose: bool,
):
    """Trailblazer - Monitor analyses"""
    if verbose:
        log_format = "%(asctime)s %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s"
    else:
        log_format = "%(message)s" if sys.stdout.isatty() else None

    coloredlogs.install(level=log_level, fmt=log_format)

    context.obj = (
        ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=Path(config.name))
        if config
        else {}
    )
    context.obj["database"] = database or context.obj.get("database", "sqlite:///:memory:")
    context.obj["trailblazer"] = Store(context.obj["database"])


@base.command()
@click.option("--reset", is_flag=True, help="reset database before setting up tables")
@click.option("--force", is_flag=True, help="bypass manual confirmations")
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    existing_tables = context.obj["trailblazer"].engine.table_names()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        context.obj["trailblazer"].drop_all()
    elif existing_tables:
        LOG.warning("Database already exists, use '--reset'")
        context.abort()
    context.obj["trailblazer"].setup()
    LOG.info(f"Success! New tables: {', '.join(context.obj['trailblazer'].engine.table_names())}")


@base.command()
@click.pass_context
def scan(context):
    """Scan ongoing analyses in SLURM"""
    context.obj["trailblazer"].update_ongoing_analyses()
    LOG.info("All analyses updated!")


@base.command("update-analysis")
@click.argument("analysis_id")
@click.pass_context
def update_analysis(context, analysis_id: int):
    """Update status of a single analysis"""
    context.obj["trailblazer"].update_run_status(analysis_id=analysis_id)


@base.command("add-user")
@click.option("--name", help="Name of new user to add")
@click.argument("email", default=environ_email())
@click.pass_context
def add_user_to_db(context, email: str, name: str) -> None:
    """Add a new user to the database."""
    trailblazer_db: Store = context.obj["trailblazer"]
    existing_user = trailblazer_db.get_user(email=email, exclude_archived=False)
    if is_existing_user(user=existing_user, email=email):
        return
    new_user = trailblazer_db.add_user(email=email, name=name)
    LOG.info(f"New user added: {new_user.email} ({new_user.id})")


@base.command("get-user")
@click.argument("email", default=environ_email())
@click.pass_context
def get_user_from_db(context, email: str) -> None:
    """Display information about an existing user."""
    trailblazer_db: Store = context.obj["trailblazer"]
    existing_user = trailblazer_db.get_user(email=email, exclude_archived=False)
    if not is_existing_user(user=existing_user, email=email):
        return
    LOG.info(f"Existing user found: {existing_user.to_dict()}")


@base.command("get-users")
@click.option("--name", type=click.types.STRING, help="Name of new users to list")
@click.option("--email", type=click.types.STRING, help="Name of new users to list")
@click.option("--exclude-archived", is_flag=True, help="Exclude archived users")
@click.pass_context
def get_users_from_db(context, name: str, email: str, exclude_archived: bool) -> None:
    """Display information about existing users."""
    trailblazer_db: Store = context.obj["trailblazer"]
    users: List[User] = trailblazer_db.get_users(
        email=email, exclude_archived=exclude_archived, name=name
    )
    LOG.info("Listing users in database:")
    for user in users:
        LOG.info(f"{user.to_dict()}")


@base.command("archive-user")
@click.argument("email", default=environ_email())
@click.pass_context
def archive_user(context, email: str) -> None:
    """Archive an existing user identified by email."""
    trailblazer_db: Store = context.obj["trailblazer"]
    existing_user: User = trailblazer_db.get_user(email=email, exclude_archived=False)

    if not is_existing_user(user=existing_user, email=email):
        return
    if is_user_archived(user=existing_user, email=email):
        return

    trailblazer_db.update_user_is_archived(user=existing_user, archive=True)
    LOG.info(f"User archived: {email}")


@base.command("unarchive-user")
@click.argument("email", default=environ_email())
@click.pass_context
def unarchive_user(context, email: str) -> None:
    """Unarchive an existing user identified by email."""
    trailblazer_db: Store = context.obj["trailblazer"]
    existing_user: User = trailblazer_db.get_user(email=email, exclude_archived=False)

    if not is_existing_user(user=existing_user, email=email):
        return
    if not is_user_archived(user=existing_user, email=email):
        return

    trailblazer_db.update_user_is_archived(user=existing_user, archive=False)
    LOG.info(f"User unarchived: {email}")


@base.command()
@click.argument("analysis_id", type=int)
@click.pass_context
def cancel(context, analysis_id):
    """Cancel all jobs in a run."""
    try:
        context.obj["trailblazer"].cancel_analysis(analysis_id=analysis_id, email=environ_email())
    except Exception as e:
        LOG.error(e)


@base.command("set-completed")
@click.argument("analysis_id", type=int)
@click.pass_context
def set_analysis_completed(context, analysis_id):
    """Set status of an analysis to "COMPLETED" """
    try:
        context.obj["trailblazer"].set_analysis_completed(analysis_id=analysis_id)
    except Exception as e:
        LOG.error(e)


@base.command("set-status")
@click.option(
    "--status",
    is_flag=False,
    type=str,
    help=f"Status to be set. Can take the values:{STATUS_OPTIONS}",
)
@click.argument("case_id", type=str)
@click.pass_context
def set_analysis_status(
    context,
    case_id: str,
    status: str,
):
    """Set the status of the latest analysis for a given CASE_ID."""
    try:
        context.obj["trailblazer"].set_analysis_status(case_id=case_id, status=status)
    except ValueError as e:
        LOG.error(e)
        raise click.Abort from e
    except Exception as e:
        LOG.error(e)


@base.command()
@click.option("--force", is_flag=True, help="Force delete if analysis ongoing")
@click.option("--cancel-jobs", is_flag=True, help="Cancel all ongoing jobs before deleting")
@click.argument("analysis_id", type=int)
@click.pass_context
def delete(context, analysis_id: int, force: bool, cancel_jobs: bool):
    """Delete analysis completely from database, and optionally cancel all ongoing jobs"""
    try:
        if cancel_jobs:
            context.obj["trailblazer"].cancel_analysis(analysis_id=analysis_id)

        context.obj["trailblazer"].delete_analysis(analysis_id=analysis_id, force=force)
    except Exception as e:
        LOG.error(e)


@base.command("ls")
@click.option(
    "-s", "--status", type=click.Choice(STATUS_OPTIONS), help="Find analysis with specified status"
)
@click.option("-b", "--before", help="Find analyses started before date")
@click.option("-c", "--comment", help="Find analysis with comment")
@click.pass_context
def ls_cmd(context, before, status, comment):
    """Display recent logs for analyses."""
    runs = (
        context.obj["trailblazer"]
        .analyses(
            status=status,
            deleted=False,
            before=parse_date(before) if before else None,
            comment=comment,
        )
        .limit(30)
    )
    for run_obj in runs:
        if run_obj.status == "pending":
            message = f"{run_obj.id} | {run_obj.family} [{run_obj.status.upper()}]"
        else:
            message = (
                f"{run_obj.id} | {run_obj.family} {run_obj.started_at.date()} "
                f"[{run_obj.type.upper()}/{run_obj.status.upper()}]"
            )
            if run_obj.status == "running":
                message = click.style(f"{message} - {run_obj.progress * 100}/100", fg="blue")
            elif run_obj.status == "completed":
                message = click.style(f"{message} - {run_obj.completed_at}", fg="green")
            elif run_obj.status == "failed":
                message = click.style(message, fg="red")
        click.echo(message)



from typing import List, Dict, Optional, Callable

from trailblazer.apps.slurm.models import SqueueResult
from trailblazer.constants import FileFormat, SlurmJobStatus, TrailblazerStatus
from trailblazer.exc import EmptySqueueError
from trailblazer.io.controller import ReadStream
from trailblazer.apps.slurm.utils import formatters


def get_squeue_result(squeue_response: str) -> SqueueResult:
    """Return SqueueResult object from squeue response.
    Raises:
        TrailblazerError: when no entries were returned by squeue command.
    """
    if not squeue_response:
        raise EmptySqueueError("No jobs found in SLURM registry")
    squeue_response_content: List[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_response,
        read_to_dict=True,
    )
    return SqueueResult(jobs=squeue_response_content)


def reformat_squeue_result_job_step(data_analysis: str, job_step: str) -> str:
    """Standardise job step string according to data analysis."""
    formatter: Callable = formatters.formatter_map.get(data_analysis, formatters.reformat_undefined)
    return formatter(job_step=job_step)


def _get_analysis_single_status(jobs_status_distribution: Dict[str, float]) -> str:
    """Return true if only one status in jobs statuses."""
    if len(jobs_status_distribution) == 1:
        single_status: str = jobs_status_distribution.popitem()[0]
        return (
            TrailblazerStatus.FAILED
            if single_status == SlurmJobStatus.TIME_OUT
            else TrailblazerStatus[single_status.upper()]
        )


def _is_analysis_failed(jobs_status_distribution: Dict[str, float]) -> bool:
    """Return true if any job was broken."""
    broken_statuses: List[str] = [SlurmJobStatus.FAILED, SlurmJobStatus.TIME_OUT]
    for broken_status in broken_statuses:
        if jobs_status_distribution.get(broken_status):
            return True


def _is_analysis_ongoing(jobs_status_distribution: Dict[str, float]) -> bool:
    """Return True if analysis is still ongoing."""
    return bool(
        jobs_status_distribution.get(SlurmJobStatus.RUNNING)
        or jobs_status_distribution.get(SlurmJobStatus.PENDING)
    )


def get_current_analysis_status(jobs_status_distribution: Dict[str, float]) -> str:
    """Return current analysis status based on jobs status distribution."""
    single_analysis_status: Optional[str] = _get_analysis_single_status(
        jobs_status_distribution=jobs_status_distribution
    )
    if single_analysis_status:
        return single_analysis_status
    is_analysis_ongoing: bool = _is_analysis_ongoing(
        jobs_status_distribution=jobs_status_distribution
    )
    is_analysis_failed: bool = _is_analysis_failed(
        jobs_status_distribution=jobs_status_distribution
    )
    if is_analysis_failed:
        return TrailblazerStatus.ERROR if is_analysis_ongoing else TrailblazerStatus.FAILED
    if is_analysis_ongoing:
        return TrailblazerStatus.RUNNING
    return TrailblazerStatus.CANCELLED
